@@ -2,13 +2,12 @@ import functools
 import logging
 
 from django.conf import settings
-from django.core.cache import cache, parse_backend_uri
 from django.db import models
 from django.db.models import signals
 from django.db.models.sql import query
 from django.utils import encoding
 
-from .invalidation import invalidator, flush_key, make_key, byid
+from .invalidation import invalidator, flush_key, make_key, byid, cache
 
 
 class NullHandler(logging.Handler):
@@ -24,6 +23,8 @@ FOREVER = 0
 NO_CACHE = -1
 CACHE_PREFIX = getattr(settings, 'CACHE_PREFIX', '')
 FETCH_BY_ID = getattr(settings, 'FETCH_BY_ID', False)
+CACHE_EMPTY_QUERYSETS = getattr(settings, 'CACHE_EMPTY_QUERYSETS', False)
+TIMEOUT = getattr(settings, 'CACHE_COUNT_TIMEOUT', None)
 
 
 class CachingManager(models.Manager):
@@ -32,7 +33,7 @@ class CachingManager(models.Manager):
     use_for_related_fields = True
 
     def get_query_set(self):
-        return CachingQuerySet(self.model)
+        return CachingQuerySet(self.model, using=self._db)
 
     def contribute_to_class(self, cls, name):
         signals.post_save.connect(self.post_save, sender=cls)
@@ -69,14 +70,23 @@ class CacheMachine(object):
     called to get an iterator over some database results.
     """
 
-    def __init__(self, query_string, iter_function, timeout=None):
+    def __init__(self, query_string, iter_function, timeout=None, db='default'):
         self.query_string = query_string
         self.iter_function = iter_function
         self.timeout = timeout
+        self.db = db
 
     def query_key(self):
-        """Generate the cache key for this query."""
-        return make_key('qs:%s' % self.query_string, with_locale=False)
+        """
+        Generate the cache key for this query.
+
+        Database router info is included to avoid the scenario where related
+        cached objects from one DB (e.g. slave) are saved in another DB (e.g.
+        master), throwing a Django ValueError in the process. Django prevents
+        cross DB model saving among related objects.
+        """
+        query_db_string = u'qs:%s::db:%s' % (self.query_string, self.db)
+        return make_key(query_db_string, with_locale=False)
 
     def __iter__(self):
         try:
@@ -104,7 +114,7 @@ class CacheMachine(object):
                 to_cache.append(obj)
                 yield obj
         except StopIteration:
-            if to_cache:
+            if to_cache or CACHE_EMPTY_QUERYSETS:
                 self.cache_objects(to_cache)
             raise
 
@@ -126,7 +136,8 @@ class CachingQuerySet(models.query.QuerySet):
         return flush_key(self.query_key())
 
     def query_key(self):
-        sql, params = self.query.get_compiler(using=self.db).as_sql()
+        clone = self.query.clone()
+        sql, params = clone.get_compiler(using=self.db).as_sql()
         return sql % params
 
     def iterator(self):
@@ -141,7 +152,7 @@ class CachingQuerySet(models.query.QuerySet):
                 return iterator()
             if FETCH_BY_ID:
                 iterator = self.fetch_by_id
-            return iter(CacheMachine(query_string, iterator, self.timeout))
+            return iter(CacheMachine(query_string, iterator, self.timeout, db=self.db))
 
     def fetch_by_id(self):
         """
@@ -156,7 +167,7 @@ class CachingQuerySet(models.query.QuerySet):
         # order_by.
         vals = self.values_list('pk', *self.query.extra.keys())
         pks = [val[0] for val in vals]
-        keys = dict((byid(self.model._cache_key(pk)), pk) for pk in pks)
+        keys = dict((byid(self.model._cache_key(pk, self.db)), pk) for pk in pks)
         cached = dict((k, v) for k, v in cache.get_many(keys).items()
                       if v is not None)
 
@@ -188,13 +199,12 @@ class CachingQuerySet(models.query.QuerySet):
         return others
 
     def count(self):
-        timeout = getattr(settings, 'CACHE_COUNT_TIMEOUT', None)
         super_count = super(CachingQuerySet, self).count
         query_string = 'count:%s' % self.query_key()
-        if timeout is None:
+        if self.timeout == NO_CACHE or TIMEOUT is None:
             return super_count()
         else:
-            return cached_with(self, super_count, query_string, timeout)
+            return cached_with(self, super_count, query_string, TIMEOUT)
 
     def cache(self, timeout=None):
         qs = self._clone()
@@ -210,7 +220,7 @@ class CachingQuerySet(models.query.QuerySet):
         return qs
 
 
-class CachingMixin:
+class CachingMixin(object):
     """Inherit from this class to get caching and invalidation helpers."""
 
     def flush_key(self):
@@ -219,16 +229,16 @@ class CachingMixin:
     @property
     def cache_key(self):
         """Return a cache key based on the object's primary key."""
-        return self._cache_key(self.pk)
+        return self._cache_key(self.pk, self._state.db)
 
     @classmethod
-    def _cache_key(cls, pk):
+    def _cache_key(cls, pk, db):
         """
         Return a string that uniquely identifies the object.
 
         For the Addon class, with a pk of 2, we get "o:addons.addon:2".
         """
-        key_parts = ('o', cls._meta, pk)
+        key_parts = ('o', cls._meta, pk, db)
         return ':'.join(map(encoding.smart_unicode, key_parts))
 
     def _cache_keys(self):
@@ -236,7 +246,7 @@ class CachingMixin:
         fks = dict((f, getattr(self, f.attname)) for f in self._meta.fields
                     if isinstance(f, models.ForeignKey))
 
-        keys = [fk.rel.to._cache_key(val) for fk, val in fks.items()
+        keys = [fk.rel.to._cache_key(val, self._state.db) for fk, val in fks.items()
                 if val is not None and hasattr(fk.rel.to, '_cache_key')]
         return (self.cache_key,) + tuple(keys)
 
@@ -274,7 +284,7 @@ def cached_with(obj, f, f_key, timeout=None):
         obj_key = (obj.query_key() if hasattr(obj, 'query_key')
                    else obj.cache_key)
     except AttributeError:
-        log.warning(u'%r cannot be cached.' % obj)
+        log.warning(u'%r cannot be cached.' % encoding.smart_str(obj))
         return f()
 
     key = '%s:%s' % tuple(map(encoding.smart_str, (f_key, obj_key)))
@@ -326,8 +336,9 @@ class MethodWrapper(object):
         k = lambda o: o.cache_key if hasattr(o, 'cache_key') else o
         arg_keys = map(k, args)
         kwarg_keys = [(key, k(val)) for key, val in kwargs.items()]
-        key = 'm:%s:%s:%s:%s' % (self.obj.cache_key, self.func.__name__,
-                                 arg_keys, kwarg_keys)
+        key_parts = ('m', self.obj.cache_key, self.func.__name__,
+                     arg_keys, kwarg_keys)
+        key = ':'.join(map(encoding.smart_unicode, key_parts))
         if key not in self.cache:
             f = functools.partial(self.func, self.obj, *args, **kwargs)
             self.cache[key] = cached_with(self.obj, f, key)
