@@ -5,7 +5,8 @@ import logging
 import socket
 
 from django.conf import settings
-from django.core.cache import cache, parse_backend_uri
+from django.core.cache import cache as default_cache, get_cache, parse_backend_uri
+from django.core.cache.backends.base import InvalidCacheBackendError
 from django.utils import encoding, translation
 
 try:
@@ -13,6 +14,12 @@ try:
 except ImportError:
     redislib = None
 
+# Look for an own cache first before falling back to the default cache
+try:
+    cache = get_cache('cache_machine')
+except (InvalidCacheBackendError, ValueError):
+    cache = default_cache
+
 
 CACHE_PREFIX = getattr(settings, 'CACHE_PREFIX', '')
 FETCH_BY_ID = getattr(settings, 'FETCH_BY_ID', False)
@@ -38,7 +45,7 @@ def flush_key(obj):
 
 
 def byid(obj):
-    key = obj if isinstance(obj, basestring) else obj._cache_key(obj.pk)
+    key = obj if isinstance(obj, basestring) else obj.cache_key
     return make_key('byid:' + key)
 
 
@@ -64,7 +71,6 @@ def safe_redis(return_type):
     return decorator
 
 
-
 class Invalidator(object):
 
     def invalidate_keys(self, keys):
@@ -74,7 +80,7 @@ class Invalidator(object):
         flush, flush_keys = self.find_flush_lists(keys)
 
         if flush:
-            cache.set_many(dict((k, None) for k in flush), 5)
+            cache.delete_many(flush)
         if flush_keys:
             self.clear_flush_lists(flush_keys)
 
